#
# Makefile for the kernel NVRAM block device drivers.
#
# Advantech eAutomation Division
#


UNAME := $(shell uname -r)
PWD := $(shell pwd)
ADVMOD := advnvram

# FSTYPE 
# 1. Leave it empty
#    It can be failed if NVRAM does not contain a file system.
# 2. File system types, such as ext2
#    It will force to make a specified file system.
#    Everything in NVRAM will be destroied.
# FSTYPE :=
# FSTYPE := ext2
# FSTYPE := vfat
FSTYPE := ext2
ifneq ($(FSTYPE),) 
MKFS := @mkfs -t $(FSTYPE) /dev/$(ADVMOD) > /dev/null
endif
ifeq "$(UNAME)" "2.4.27-2-386"
LINUX_SRC=/usr/src/kernel-headers-2.4.27-2-386
else 
LINUX_SRC=	/usr/src/linux-2.4
endif
ARCH_FLAGS=    $(I386_FLAGS)

# Define the advnvram driver options as appropriate .
INCL_FLAGS=  -I. -I$(LINUX_SRC)/include
DEF_FLAGS=   -D__KERNEL__ -DLINUX -DLOCAL_HEADERS -DEXPORT_SYMTAB 
I386_FLAGS=  -march=i686 -O2 -Wall 
CFLAGS=	     $(ARCH_FLAGS) $(INCL_FLAGS) $(DEF_FLAGS) -DMODULE
CC=          gcc

default: advnvram.o
	ld -r  advnvram_init.o advnvram_pci.o advnvram_fops.o advnvram_request.o -o advnvram.o

	
advnvram.o:advnvram_init.o advnvram_pci.o advnvram_fops.o advnvram_request.o
	

install:
	$(shell if grep $(ADVMOD) /proc/mounts > /dev/null ; then \
	umount /mnt/$(ADVMOD) ; fi)
	$(shell if grep $(ADVMOD) /proc/modules > /dev/null ; then \
	rmmod $(ADVMOD) ; fi)
	$(shell [ -b /dev/$(ADVMOD) ] && rm -f /dev/$(ADVMOD))
	@insmod $(ADVMOD).o
	

	
uninstall:
	$(shell if grep $(ADVMOD) /proc/mounts > /dev/null ; then \
	 umount /mnt/$(ADVMOD) ; fi)
	$(shell if grep $(ADVMOD) /proc/modules > /dev/null ; then \
	 rmmod $(ADVMOD) ; fi)

nodes:install 
	$(shell [ -b /dev/advnvram0 ] && rm -f /dev/advnvram0)
	$(shell [ -d /mnt/advnvram0 ] && \
		rmdir --ignore-fail-on-non-empty /mnt/advnvram0)
	@mknod /dev/advnvram0 b $$(awk '$$2=="advnvram" {print $$1}' /proc/devices ) 0
	@chmod a+wx /dev/advnvram0
	$(shell [ -b /dev/advnvram1 ] && rm -f /dev/advnvram1)
	$(shell [ -d /mnt/advnvram1 ] && \
		rmdir --ignore-fail-on-non-empty /mnt/advnvram1)
	@mknod /dev/advnvram1 b $$(awk '$$2=="advnvram" {print $$1}' /proc/devices ) 1
	@chmod a+wx /dev/advnvram1
	$(shell [ -b /dev/advnvram2 ] && rm -f /dev/advnvram2)
	$(shell [ -d /mnt/advnvram2 ] && \
		rmdir --ignore-fail-on-non-empty /mnt/advnvram2)
	@mknod /dev/advnvram2 b $$(awk '$$2=="advnvram" {print $$1}' /proc/devices ) 2
	@chmod a+wx /dev/advnvram2
	$(shell [ -b /dev/advnvram3 ] && rm -f /dev/advnvram3)
	$(shell [ -d /mnt/advnvram3 ] && \
		rmdir --ignore-fail-on-non-empty /mnt/advnvram3)
	@mknod /dev/advnvram3 b $$(awk '$$2=="advnvram" {print $$1}' /proc/devices ) 3
	@chmod a+wx /dev/advnvram3	

clean:
	@rm -f *.o *~

