#ifndef _ADVNVRAM_H
#define _ADVNVRAM_H

#include <linux/fs.h>
#include <linux/genhd.h>
#include <asm/uaccess.h>
#include <asm/io.h>




#define ADVANTECH_NVRAM_VER	"2.18"
#define ADVANTECH_NVRAM_DATE	"06/25/2013"

#define DEVICE_NR(device) MINOR(device)   
#define DEVICE_NAME "advnvram"       // name for messaging: Advantech NVRAMDisk
//#define DEVICE_INTR adsd_intrptr    // pointer to the bottom half
#define DEVICE_NO_RANDOM            // no entropy to contribute

#include <linux/version.h>
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
#include <linux/blkdev.h>
#else
#include <linux/blk.h>
#endif

#include <linux/blkpg.h>

#ifdef ADVANTECH_DEBUG
#define DEBUGPRINT	printk
#else
#define DEBUGPRINT(a, ...)
#endif

#define ADVANTECH_INFO
#ifdef ADVANTECH_INFO
#define INFOPRINT	printk
#else
#define INFOPRINT(a, ...)
#endif

// define the register 
#define ADVANTECH_NVRAM_MANUFACTURE_ID   0xAA
#define DEFAULT_BANK_MEMORY_ADDRESS     0xDC000
#define DEFAULT_BANK_SWITCH_ADDRESS     0x1E0
#define DEFAULT_BANK_ENABLE_ADDRESS     0x1E1
#define DEFAULT_MANUFACTUREID_ADDRESS   0x1E6
#define DEFAULT_BANK_NUMBER_ADDRESS     0x1E7
#define DEFAULT_BANK_LENGTH             0x4000       // 16 KB
#define DEFAULT_BANK_NUMBER             0x20
#define DEFAULT_DISK_SIZE               (512*1024)   // 512 KB
#define DEFAULT_ROOT_DIR_ENTRIES        512
#define DEFAULT_SECTORS_PER_CLUSTER     2

#define ADVNVRAM_MAJOR                 0                       // dynamic major
#define NUM_NVRAMDISKS                 (8)
#define ADVNVRAM_CHUNK1024(x)          (( x ) >> 10 )          // chunk to KBytes
#define MAX_DISK 16;
#define TRUE                          (1)
#define FALSE                         (0)

typedef struct _DISK_INFO {
   u_long   ul_DiskSize;
   u_long   ul_BankMemoryAddress;
   u_long   ul_BankSwitchAddress;
   u_long   ul_BankEnableAddress;
   u_long   ul_ManufactureIDAddress;
   u_long   ul_BankNumberAddress;
   u_long   ul_BankLength;
   u_long   ul_BankNumber;
   u_long   ul_BankForceFormat;
   u_long   ul_BankMappedMemoryAddress;
   u_long   ul_PureDRAM;
   u_long   ul_IsPci;
   u_long   ul_FPGAVersion;
   unsigned char * uc_ptaddr;		  // pointer to the remap space 
} DISK_INFO, *PDISK_INFO;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
extern int asd_hardsec[];
extern int asd_blocksizes[];
extern int asd_sizes[];
#endif

/**
 * The simulated PCM device with a RAM disk backing store
 */
struct advnvram_device {
	
	/// The device number
	int advnvram_index;

	/// The backing data store
	DISK_INFO   *  advnvram_info;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	/// The lock protecting the data store
	spinlock_t advnvram_lock;

	/// Request queue
	struct request_queue* advnvram_queue;
	/// Disk
	struct gendisk* advnvram_disk;	
#endif

	/// The collection of lists the device belongs to
	struct list_head advnvram_list;

};

extern int     advsd_major; 
extern DISK_INFO AdvNVRAMInfo[];
extern int advnvram_num_devices;
extern struct advnvram_device *g_advnvram;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)	
extern int advnvram_open ( struct inode *inode, struct file *filp );
#else
extern int advnvram_open(struct block_device *bdev, fmode_t mode);
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29)
extern int advnvram_ioctl(struct block_device *bdev, fmode_t mode,unsigned int cmd, unsigned long arg);
#else
extern int advnvram_ioctl(struct inode *i, struct file *f, unsigned int cmd, unsigned long arg);
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
extern int advnvram_release ( struct inode *inode, struct file *filp );
#else
extern int advnvram_release(struct gendisk *disk, fmode_t mode);
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,23)
extern  void advnvram_do_request( struct request_queue * q );
#else
extern  void advnvram_do_request( request_queue_t * q );
#endif
extern int init_advnvram_pci(void);
extern void exit_advnvram_pci(void);


#endif
