#include <linux/version.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/ioport.h>
#include <linux/hdreg.h>  /* HDIO_GETGEO */
#include <linux/fs.h>
#include <linux/genhd.h>
#include <asm/uaccess.h>


#include <asm/io.h>

#include "advnvram.h"

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,28)	
int advnvram_open ( struct inode *inode, struct file *filp )
#else
int advnvram_open(struct block_device *bdev, fmode_t mode)
#endif
{
	//int device;    
	int rc = -ENODEV;    
	unsigned char uc_temp;
	unsigned long size, paddr;	
	DISK_INFO   *  advnvram_info = NULL;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)	
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,28)	
	int minor = 0;
	minor = iminor(inode); 
	advnvram_info  = &AdvNVRAMInfo[minor];
#else
	struct advnvram_device *advnvram = NULL;
	advnvram = (struct advnvram_device *)(bdev->bd_disk->private_data);
	advnvram_info = advnvram->advnvram_info;
#endif
#else
	int minor = 0;
	minor = DEVICE_NR( inode->i_rdev );
	advnvram_info  = &AdvNVRAMInfo[minor];
#endif
	if (advnvram_info->uc_ptaddr == NULL )
	{
		return -ENODEV;
	}
	
	
	paddr = advnvram_info->ul_BankMemoryAddress;
	size  = advnvram_info->ul_BankLength  ;
	DEBUGPRINT( DEVICE_NAME": Memory Address is: %lx, Size: %lx\n", paddr, size);		
 
	
	DEBUGPRINT( DEVICE_NAME": vaddr = %p\n", advnvram_info->uc_ptaddr );		
	
	if ( advnvram_info->ul_IsPci == 0 )
	{
		// reset bank number
		outb( 0, advnvram_info->ul_BankSwitchAddress);  
		// enable NVRAM R/W capability  
		outb( 1, advnvram_info->ul_BankEnableAddress);
	}
	// check AdvNVRAMDisk connected or not
	uc_temp = advnvram_info->uc_ptaddr[ 0 ];
	advnvram_info->uc_ptaddr[ 0 ] = ~(advnvram_info->uc_ptaddr[ 0 ]);
	DEBUGPRINT( DEVICE_NAME": uc_temp = %x, uc_ptaddr[0] = %x \n", uc_temp, advnvram_info->uc_ptaddr[ 0 ]);
	
	if( advnvram_info->uc_ptaddr[ 0 ] != (unsigned char) (~uc_temp)) 
	{
		//INFOPRINT( DEVICE_NAME": AdvNVRAMDisk doesn't connect yet! \n");
		if ( advnvram_info->ul_IsPci == 0 )
		{
			outb(0, advnvram_info->ul_BankEnableAddress);   
		}
		iounmap(( void* )advnvram_info->uc_ptaddr );      
		rc = -ENODEV;
	}
	advnvram_info->uc_ptaddr[ 0 ] = uc_temp; 
	 
	if ( advnvram_info->ul_IsPci == 0 )
	{
		outb( 0, advnvram_info->ul_BankSwitchAddress );
		outb( 0, advnvram_info->ul_BankEnableAddress );
	}
	return 0;
	
//err_out:    
//	return rc;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,28)
int advnvram_ioctl(struct block_device *bdev, fmode_t mode,unsigned int cmd, unsigned long arg)
#else
int advnvram_ioctl(struct inode *i, struct file *f, unsigned int cmd, unsigned long arg)
#endif
{
   struct hd_geometry geo;
   DISK_INFO   *  advnvram_info = NULL;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,28)

   struct advnvram_device *advnvram = NULL;
  
   advnvram = (struct advnvram_device *)(bdev->bd_disk->private_data);
   advnvram_info = advnvram->advnvram_info;
#else
   int minor = 0;
   minor = iminor(i); 
   advnvram_info  = &AdvNVRAMInfo[minor];
#endif 
#else
   int minor = 0;
   minor = DEVICE_NR( i->i_rdev );
   advnvram_info  = &AdvNVRAMInfo[minor];
#endif
   //if (cmd == HDIO_GETGEO) {

   switch ( cmd )
   {
	case HDIO_GETGEO:
		geo.heads     = 32;
		geo.sectors   = 2;
		geo.cylinders = ( (advnvram_info->ul_DiskSize) >> 9) / (geo.heads * geo.sectors);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		geo.start     = 0;

		if ( copy_to_user((void* /*__user **/)arg, &geo, sizeof(geo)))

#else
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,28)
		geo.start     = get_start_sect(bdev);
#else
		geo.start     = get_start_sect(i->i_bdev);
#endif    
		if (copy_to_user((void __user *) arg, &geo, sizeof(geo)))
#endif
			return -EFAULT;
		break;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	case BLKGETSIZE:
		if (!arg)
			break;
		if ( put_user( asd_sizes[minor] <<1, (unsigned long *)arg ))
			return -EFAULT;
		break;
	case BLKGETSIZE64:
		if ( put_user((u64)(asd_sizes[minor]) << 10, (u64*)arg))
			return -EFAULT;
		break;
	case BLKROSET:
	case BLKROGET:
	case BLKSSZGET:
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,28)
		if ( blk_ioctl( bdev, mode,cmd, arg ))
#else
		if ( blk_ioctl( i->i_rdev, cmd, arg ))
#endif
			return -EFAULT;
		break;
#endif
	}
//	return -EINVAL;
	return 0;
}

//=============================================================================
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,28)
int advnvram_release ( struct inode *inode, struct file *filp )
#else
int advnvram_release(struct gendisk *disk, fmode_t mode)
#endif
{
   
	/*
	* FIXME: unmap memory
	*/
	
	return 0;
}

