#include <linux/version.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/ioport.h>
#include <linux/hdreg.h>  
#include <asm/uaccess.h>
#include <asm/io.h>
#include <linux/vmalloc.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/blkpg.h>


#include "advnvram.h"

int advnvram_num_devices = 0;
//int advsd_major;
DISK_INFO AdvNVRAMInfo[NUM_NVRAMDISKS] = {{0,},};
// The list of AdvNVRAM devices
static LIST_HEAD(advnvram_devices);

struct advnvram_device* advnvram_alloc(int index);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
static int asd_hardsec[ NUM_NVRAMDISKS ];      // Size of real blocks in byte
static int asd_blocksizes[ NUM_NVRAMDISKS ];   // Size of 1024 byte blocks
static int asd_sizes[ NUM_NVRAMDISKS ];        // Size of the device in bytes
#endif
// The block device operations defined in fs.h
static struct block_device_operations advnvram_fops =
{
	owner:		THIS_MODULE,   
	open:		advnvram_open,
	release:	advnvram_release,
	ioctl:		advnvram_ioctl,
};

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)

//The mutex guarding the list of devices
static DEFINE_MUTEX(advnvram_devices_mutex);

/**
 * Initialize one device
 */
static struct advnvram_device *advnvram_init_one(int i)
{
	struct advnvram_device *advnvram;

	list_for_each_entry(advnvram, &advnvram_devices, advnvram_list) {
		if (advnvram->advnvram_index == i)
			goto out;
	}

	advnvram = advnvram_alloc(i);
	if (advnvram) {
		add_disk(advnvram->advnvram_disk);
		list_add_tail(&advnvram->advnvram_list, &advnvram_devices);
	}
out:
	return advnvram;
}


/**
 * Probe a device
 */
static struct kobject *advnvram_probe(dev_t dev, int *part, void *data)
{
	struct advnvram_device *advnvram;
	struct kobject *kobj;

	mutex_lock(&advnvram_devices_mutex);
	advnvram = advnvram_init_one(dev & MINORMASK);
	kobj = advnvram ? get_disk(advnvram->advnvram_disk) : ERR_PTR(-ENOMEM);
	mutex_unlock(&advnvram_devices_mutex);

	*part = 0;
	return kobj;
}
#endif 


/**
 * Allocate the NVRAMDisk device
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
struct advnvram_device* advnvram_alloc(int index)
{
	struct advnvram_device *advnvram;
	struct gendisk *disk;


	// Allocate the device
	advnvram = kzalloc(sizeof(*advnvram), GFP_KERNEL);
	if (!advnvram) 
		goto out;
	advnvram->advnvram_index = index;
	advnvram->advnvram_info = &AdvNVRAMInfo[index];
	spin_lock_init(&advnvram->advnvram_lock);
	
	advnvram->advnvram_queue = blk_init_queue(advnvram_do_request, &advnvram->advnvram_lock);
	if (!advnvram->advnvram_queue)
	{
		INFOPRINT( DEVICE_NAME ": Error to initialize queue!\n");
		goto out_free_dev;
	}
	

	// Allocate the disk device
	disk = advnvram->advnvram_disk = alloc_disk(1 /* cannot be partitioned */);
	if (!disk) 
		goto out_free_queue;
	disk->major	       = advsd_major;
	disk->first_minor  = index;
	disk->fops         = &advnvram_fops;
	disk->private_data = advnvram;
	disk->queue	       = advnvram->advnvram_queue;
	disk->flags       |= GENHD_FL_SUPPRESS_PARTITION_INFO;
	sprintf(disk->disk_name, "advnvram%d", index);
	
	advnvram->advnvram_info->ul_DiskSize  = advnvram->advnvram_info->ul_BankLength  * advnvram->advnvram_info->ul_BankNumber;
	advnvram->advnvram_info->uc_ptaddr = ioremap_nocache( advnvram->advnvram_info->ul_BankMemoryAddress, advnvram->advnvram_info->ul_BankLength );
	INFOPRINT( "%s: Advantech NVRAM Disk %ld KBytes", 
			disk->disk_name,
			ADVNVRAM_CHUNK1024(advnvram->advnvram_info->ul_DiskSize ));
	if ( advnvram->advnvram_info->ul_IsPci == 1 ) 
	{
		INFOPRINT( ", FPGA Version: %04x\n",(unsigned int)advnvram->advnvram_info->ul_FPGAVersion);
	}
	else
	{
		INFOPRINT( "\n" );
	}

	set_capacity(disk, advnvram->advnvram_info->ul_DiskSize >> 9); // 512 bytes per sector
	
	return advnvram;
	// Cleanup on error

out_free_queue:
	blk_cleanup_queue(advnvram->advnvram_queue);
out_free_dev:
	kfree(advnvram);
out:
	return NULL;
}
#else  //2.4 kernel
struct advnvram_device* advnvram_alloc(int index)
{
	struct advnvram_device *advnvram;
	
	// Allocate the device
	advnvram = kmalloc(sizeof(*advnvram), GFP_KERNEL);
	if (!advnvram) 
		goto out;
	advnvram->advnvram_index = index;
	advnvram->advnvram_info = &AdvNVRAMInfo[index];

	advnvram->advnvram_info->ul_DiskSize  = advnvram->advnvram_info->ul_BankLength  * advnvram->advnvram_info->ul_BankNumber;
	
	asd_blocksizes[index] = 1024;
	asd_hardsec[index]    = 512;
	asd_sizes[index]      = ADVNVRAM_CHUNK1024(AdvNVRAMInfo[index].ul_DiskSize) ;
	AdvNVRAMInfo[index].uc_ptaddr = ioremap_nocache( AdvNVRAMInfo[index].ul_BankMemoryAddress, AdvNVRAMInfo[index].ul_BankLength );

	INFOPRINT( "advnvram%d: Advantech NVRAM Disk %ld KBytes", 
			index,
			ADVNVRAM_CHUNK1024(AdvNVRAMInfo[index].ul_DiskSize ));
	if ( AdvNVRAMInfo[index].ul_IsPci == 1 ) 
	{
		INFOPRINT( ", FPGA Version: %04x\n",(unsigned int)advnvram->advnvram_info->ul_FPGAVersion);
	}
	else
	{
		INFOPRINT( "\n" );
	}
	return advnvram;
out:
	return NULL;
}
#endif

int init_advnvram_isa(void)
{
	   int    i_iobase, i_iorange,ret;
	   unsigned int i_ManuID, i_BankSwitch, i_BankNum;
	   unsigned char uc_temp;
	   i_iobase  = 0x1E0;
	   i_iorange = 8;
	   AdvNVRAMInfo[advnvram_num_devices].ul_BankEnableAddress    = DEFAULT_BANK_ENABLE_ADDRESS;
	   AdvNVRAMInfo[advnvram_num_devices].ul_BankLength           = DEFAULT_BANK_LENGTH;
	   AdvNVRAMInfo[advnvram_num_devices].ul_BankMemoryAddress    = DEFAULT_BANK_MEMORY_ADDRESS;
	   AdvNVRAMInfo[advnvram_num_devices].ul_BankNumber           = DEFAULT_BANK_NUMBER;
	   AdvNVRAMInfo[advnvram_num_devices].ul_BankNumberAddress    = DEFAULT_BANK_NUMBER_ADDRESS;
	   AdvNVRAMInfo[advnvram_num_devices].ul_BankSwitchAddress    = DEFAULT_BANK_SWITCH_ADDRESS;
	   AdvNVRAMInfo[advnvram_num_devices].ul_ManufactureIDAddress = DEFAULT_MANUFACTUREID_ADDRESS;
	
	   INFOPRINT( DEVICE_NAME ": Major Device Number is %d\n", advsd_major );
	
	   // request resources =====
	   if( !request_region( i_iobase, i_iorange, "Advantech NVRAMDisk" ))
	   {
		   ret = -EBUSY;
		   INFOPRINT( "Cannot request IO region!\n");	
		   goto out_blkdev;
	   }
	       
	   i_ManuID = inb( i_iobase + 6 );
	   i_BankSwitch = inb( i_iobase );    
	
	   if  ( i_ManuID == 0xAA )
	   {
		   DEBUGPRINT( DEVICE_NAME ": Manufacture ID is 0x%x\n", i_ManuID );
		   i_BankNum = inb( i_iobase + 7 ) + 1;        
		   AdvNVRAMInfo[advnvram_num_devices].ul_BankNumber = i_BankNum;
		   AdvNVRAMInfo[advnvram_num_devices].ul_DiskSize = DEFAULT_BANK_LENGTH * i_BankNum;
	   }
	   else if (advnvram_num_devices == 0 )
	   {

		   AdvNVRAMInfo[advnvram_num_devices].uc_ptaddr = ioremap_nocache( AdvNVRAMInfo[advnvram_num_devices].ul_BankMemoryAddress, AdvNVRAMInfo[advnvram_num_devices].ul_BankLength );
		   outb(0,AdvNVRAMInfo[advnvram_num_devices].ul_BankSwitchAddress); //reset bank number
		   outb(1,AdvNVRAMInfo[advnvram_num_devices].ul_BankEnableAddress); //enable NVRAM R/W capability
		   uc_temp = AdvNVRAMInfo[advnvram_num_devices].uc_ptaddr[ 0 ];
		   AdvNVRAMInfo[advnvram_num_devices].uc_ptaddr[ 0 ] = ~(AdvNVRAMInfo[advnvram_num_devices].uc_ptaddr[ 0 ]);
	
		   if( AdvNVRAMInfo[advnvram_num_devices].uc_ptaddr[ 0 ] != (unsigned char) (~uc_temp)) 
		   {
                        outb(0,AdvNVRAMInfo[advnvram_num_devices].ul_BankEnableAddress);
			iounmap(( void* )AdvNVRAMInfo[advnvram_num_devices].uc_ptaddr ); 
			goto out_blkdev;
		   }
		   iounmap(( void* )AdvNVRAMInfo[advnvram_num_devices].uc_ptaddr ); 
		   DEBUGPRINT( DEVICE_NAME ": Unknown Manufacture ID 0x%x\n", i_ManuID );
		   AdvNVRAMInfo[advnvram_num_devices].ul_DiskSize  = DEFAULT_BANK_LENGTH * DEFAULT_BANK_NUMBER;

	   }
	   else
	   {
		   goto out_blkdev;
	   }
	   release_region(i_iobase, i_iorange);
	   AdvNVRAMInfo[advnvram_num_devices].ul_IsPci  = 0;
	   advnvram_num_devices++;
           return 0;
out_blkdev:
	release_region(i_iobase, i_iorange);
	return ret;
}

/**
 * Free a ADVNVRAM device
 */
void advnvram_free(struct advnvram_device *advnvram)
{
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
	put_disk(advnvram->advnvram_disk);
	blk_cleanup_queue(advnvram->advnvram_queue);
#endif
	kfree(advnvram);

}


static int __init advnvram_init ( void )
{
	struct advnvram_device *advnvram;

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
	struct advnvram_device *next;
#endif
	int i =0;

	INFOPRINT("\n=========================================================\n");
	INFOPRINT("Advantech NVRAM Disk Driver. V%s [%s]\n",
		ADVANTECH_NVRAM_VER, ADVANTECH_NVRAM_DATE);
	INFOPRINT("Supports: Linux block driver for NVRAM\n");
	INFOPRINT("Advantech eAutomation Division.\n"); 
	INFOPRINT("=========================================================\n");
	
	// Register major, and accept a dynamic number	
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)	
	advsd_major = register_blkdev(0, DEVICE_NAME);
#else	
	advsd_major = register_blkdev(0, DEVICE_NAME,&advnvram_fops);
	
	blk_init_queue( BLK_DEFAULT_QUEUE( advsd_major ), advnvram_do_request );
	hardsect_size[advsd_major]  = asd_hardsec;
	blksize_size[ advsd_major] = asd_blocksizes;
	blk_size[ advsd_major ]     = asd_sizes;   // in KB
#endif
	if ( advsd_major < 0 ) 
	{
		INFOPRINT( DEVICE_NAME ": can't get major\n");
		return -EIO;
	}
	
   	init_advnvram_pci();
	init_advnvram_isa();
	if (advnvram_num_devices == 0)
	{
		INFOPRINT("No Advantech NVRAMDisk Found\n");
		goto out;
	}


	for (i = 0; i < advnvram_num_devices; i++) 
	{
		advnvram = advnvram_alloc(i);
		if (!advnvram)
		{
			printk("Init advnvram%d failed\n", i);
			goto out_free;
		}
		list_add_tail(&advnvram->advnvram_list, &advnvram_devices);
	}
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)	
	// Register the block devices
	list_for_each_entry(advnvram, &advnvram_devices, advnvram_list) {
		add_disk(advnvram->advnvram_disk);
	}

	//range = pcmsim_num_devices ? pcmsim_num_devices : 1UL << (MINORBITS - 1);
	blk_register_region(MKDEV(advsd_major, 0), advnvram_num_devices, THIS_MODULE, 		advnvram_probe, NULL, NULL);
#endif

	return 0;


	// Cleanup on error: Free all devices

out_free:
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
	list_for_each_entry_safe(advnvram, next, &advnvram_devices, advnvram_list) {
		list_del(&advnvram->advnvram_list);
		advnvram_free(advnvram);
	}
#endif
out:
	exit_advnvram_pci();
	unregister_blkdev(advsd_major, "advnvram");

	return -ENODEV;;


}


/**
 * Delete a device
 */
static void advnvram_del_one(struct advnvram_device *advnvram)
{
	/*if ( advnvram->advnvram_info->ul_IsPci == 1 ) 
	{
		release_mem_region(advnvram->advnvram_info->ul_BankMemoryAddress, advnvram->advnvram_info->ul_BankLength);
	}
        */
	iounmap(( void* )advnvram->advnvram_info->uc_ptaddr ); 

	
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
	list_del(&advnvram->advnvram_list);
	del_gendisk(advnvram->advnvram_disk);
	advnvram_free(advnvram);
#endif

} 

/**
 * Deinitalize a module
 */
static void __exit advnvram_exit(void)
{
	//unsigned long range;
	struct advnvram_device *advnvram;

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
	struct advnvram_device *next;
#endif

	exit_advnvram_pci();
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
	list_for_each_entry_safe(advnvram, next, &advnvram_devices, advnvram_list) {
		advnvram_del_one(advnvram);
	}

	blk_unregister_region(MKDEV(advsd_major, 0), advnvram_num_devices);
#else
	
	blk_cleanup_queue( BLK_DEFAULT_QUEUE( advsd_major));
	list_for_each_entry(advnvram, &advnvram_devices, advnvram_list) {
		advnvram_del_one(advnvram);
	}

#endif	
	unregister_blkdev(advsd_major, "advnvram");
	printk("advnvram module exit\n");
}

module_init(advnvram_init);
module_exit(advnvram_exit); 
MODULE_LICENSE("GPL");
