// #############################################################################
// *****************************************************************************
//                  Copyright ( c )2009 ~ 2013, Advantech Automation Corp.
//      THIS IS AN UNPUBLISHED WORK CONTAINING CONFIDENTIAL AND PROPRIETARY
//               INFORMATION WHICH IS THE PROPERTY OF ADVANTECH AUTOMATION CORP.
//
//    ANY DISCLOSURE, USE, OR REPRODUCTION, WITHOUT WRITTEN AUTHORIZATION FROM
//               ADVANTECH AUTOMATION CORP., IS STRICTLY PROHIBITED.
// *****************************************************************************
// #############################################################################
/***
 * File:              advnvram_pci.c
 * Author:            James Dai
 * Created:           2009-8-20
 * Revision:          1.00 
 *
 * Description:
 *
 * Driver init section for advantech PCI NVRAMDisk Disk
 *
 */

#include <linux/version.h>
#include <linux/pci.h>
#if LINUX_VERSION_CODE > KERNEL_VERSION(3,2,0)
#include <linux/export.h>
#endif
#include "advnvram.h"

#define ADVANTECH_VANDORID 0x13fe
extern int    i_iobase;
extern int    i_iorange;
static struct pci_device_id advnvram_board_table[] = {
	{ADVANTECH_VANDORID, 0x1011, PCI_ANY_ID, PCI_ANY_ID, 0, 0, PCI_ANY_ID},
	{0,}			
};

static int advnvram_pci_init_one(struct pci_dev *pdev, const struct pci_device_id *id);
static void advnvram_pci_remove_one(struct pci_dev *dev);

static struct pci_driver advnvram_driver = {
	name:		"AdvNVRAMDisk",
	id_table:	advnvram_board_table,
	probe:		advnvram_pci_init_one,
	remove:		advnvram_pci_remove_one,
};



static int advnvram_pci_init_one(struct pci_dev *pdev, const struct pci_device_id *id ) 
{
	int    i_iobase;
	int    i_iorange;
   unsigned int address,address_len;
   if (pci_enable_device (pdev))
   {
      return -EIO;
   }
   

#if 0
   printk("Get a pci can board at bus: %d, slot: %d, vendor: %x, device: %x\n", 
      pdev->bus->number,
      pdev->devfn, 
      pdev->vendor,
      pdev->device);
#endif
   address = pci_resource_start(pdev, 0);
   address_len = (pci_resource_len(pdev, 0));
   if ( request_mem_region(address, address_len , "AdvNVRAMDisk") == NULL ) 
   {
      printk ("Device I/O 0x%X is not free.\n",address);
      goto out_iobase0;
   }
   

   AdvNVRAMInfo[advnvram_num_devices].ul_BankEnableAddress    = DEFAULT_BANK_ENABLE_ADDRESS;
   AdvNVRAMInfo[advnvram_num_devices].ul_BankLength           = address_len;
   AdvNVRAMInfo[advnvram_num_devices].ul_BankMemoryAddress    = address;
   AdvNVRAMInfo[advnvram_num_devices].ul_BankNumber           = 1;
   AdvNVRAMInfo[advnvram_num_devices].ul_BankNumberAddress    = DEFAULT_BANK_NUMBER_ADDRESS;
   AdvNVRAMInfo[advnvram_num_devices].ul_BankSwitchAddress    = DEFAULT_BANK_SWITCH_ADDRESS;
   AdvNVRAMInfo[advnvram_num_devices].ul_ManufactureIDAddress = DEFAULT_MANUFACTUREID_ADDRESS;
   AdvNVRAMInfo[advnvram_num_devices].ul_IsPci = 1;
   i_iobase = pci_resource_start(pdev, 1);
   i_iorange = pci_resource_len(pdev, 1);
   AdvNVRAMInfo[advnvram_num_devices].ul_FPGAVersion = (inb(i_iobase+1) << 8) + inb(i_iobase) ;
   advnvram_num_devices++;
   return 0;
out_iobase0:
	release_mem_region(address, address_len);
	return -EIO;
}

static void advnvram_pci_remove_one( struct pci_dev *pdev) //zdd modified
{
   
   unsigned int address,address_len;
   address = pci_resource_start(pdev, 0);
   address_len = (pci_resource_len(pdev, 0));
   release_mem_region(address, address_len);
}

int init_advnvram_pci(void)
{
   int ret;
   ret = 0;
   ret = pci_register_driver(&advnvram_driver);
   if(ret < 0) 
   {
      return -ENODEV;
   }
   return 0;
}

void exit_advnvram_pci(void)
{
   pci_unregister_driver(&advnvram_driver);
}
