#include <linux/init.h>
#include <linux/module.h>
#include <linux/ioport.h>
#include <linux/hdreg.h>  /* HDIO_GETGEO */

#include <asm/uaccess.h>
#include <asm/io.h>

#define MAJOR_NR advsd_major        // force definitions on in blk.h
int     advsd_major;                // must be declared before including blk.h

#include "advnvram.h"


#define SECTOR_SHIFT 9

/**
 * Process pending requests from the queue
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,23)
void advnvram_do_request( struct request_queue * q )
#else
void advnvram_do_request( request_queue_t * q )
#endif
{
	int i, i_StartBank, i_EndBank;
	u_long start, len;
	u_long ul_StartBank, ul_EndBank, ul_DiskOffset, ul_BufOffset, ul_Size;
	unsigned char*  uc_ptDisk; 
	DISK_INFO   *  advnvram_info = NULL;

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,31)
	int err = -EIO;		
#endif  
	int rw;
	struct request *req;
	
#else
	int minor;
#endif
	
	DEBUGPRINT( "----->>>>> DO_ASD_REQUEST! \n");
	i = i_StartBank = i_EndBank = 0;   
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,31)
	req = blk_fetch_request(q);
	while ( req )
	{
		advnvram_info = ((struct advnvram_device *)req->rq_disk->private_data)->advnvram_info;
		rw = rq_data_dir(req);
		DEBUGPRINT( DEVICE_NAME": request %p: dir %d sector %ld (nr. %u) \n", 
			req, rw, (unsigned long) blk_rq_pos(req), req->current_nr_sectors );
		// per sector = 512 Bytes
		start = blk_rq_pos(req) << 9;    
		// the number of the sectors to transfer, per sector = 512 Bytes
		len  = blk_rq_cur_bytes(req); 
		err = 0; 
#else
	while ( (req = elv_next_request(q)) != NULL )
	{

		advnvram_info = ((struct advnvram_device *)req->rq_disk->private_data)->advnvram_info;
		rw = rq_data_dir(req);
		DEBUGPRINT( DEVICE_NAME": request %p: dir %d sector %ld (nr. %u) \n", 
			req, rw, (unsigned long) req->sector, req->current_nr_sectors );
		// per sector = 512 Bytes
		start = req->sector << 9;    
		// the number of the sectors to transfer, per sector = 512 Bytes
		len  = req->current_nr_sectors << 9;  
#endif
		DEBUGPRINT( DEVICE_NAME": Start addr:%ld, Length:%ld\n", start, len );   
		
		ul_StartBank = start / advnvram_info->ul_BankLength;
		ul_EndBank   = ( start + len - 1) / advnvram_info->ul_BankLength;
		DEBUGPRINT( DEVICE_NAME": Start Bank: %ld, End Bank: %ld\n", ul_StartBank, ul_EndBank );
		
		if ( ( start + len ) > advnvram_info->ul_DiskSize )
		{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,31)
			INFOPRINT( DEVICE_NAME ": bad access: block=%ld, count=%u\n",
			(long int)blk_rq_pos(req),blk_rq_cur_sectors(req));
			err = -EIO;
			goto done;
#else
			INFOPRINT( DEVICE_NAME ": bad access: block=%ld, count=%u\n",
			(unsigned long) req->sector,
			req->current_nr_sectors);
			INFOPRINT( DEVICE_NAME ": request past end of device\n");	
			end_request(req, FALSE );
			continue;
#endif
			
		}	
		
		if ( ( rw != READ ) && ( rw != WRITE ) )
		{
			//INFOPRINT( DEVICE_NAME ": unknow command: %d\n", rq_data_dir(req) );
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,31)
			err = -EIO;
			goto done;
#else
			end_request(req, FALSE );
#endif
			continue;
		}
		
		ul_BufOffset  = 0;
		ul_DiskOffset = start % advnvram_info->ul_BankLength;
		ul_Size       = ( ul_DiskOffset + len <= advnvram_info->ul_BankLength )? 
				len : (advnvram_info->ul_BankLength - ul_DiskOffset );
		for ( i = ul_StartBank; i <= ul_EndBank; i++ )
		{
			uc_ptDisk = advnvram_info->uc_ptaddr + ul_DiskOffset;
			DEBUGPRINT( DEVICE_NAME": R/W - bank = %d, ul_Size = %ld\n", i, ul_Size );
			
			// set bank
			if ( advnvram_info->ul_IsPci == 0 )
			{
				outb( i, advnvram_info->ul_BankSwitchAddress);
			}
			if ( rw == READ )
			{  
				DEBUGPRINT(DEVICE_NAME": READ-BANK = %d\n", i );
				memcpy( &req->buffer[ul_BufOffset], uc_ptDisk, ul_Size );            
			}
			else if ( rw == WRITE )
			{
				DEBUGPRINT(DEVICE_NAME": WRITE-BANK = %d\n", i );
				if ( advnvram_info->ul_IsPci == 0 )
				{
					outb( 1, advnvram_info->ul_BankEnableAddress);
				}
				memcpy( uc_ptDisk, &req->buffer[ul_BufOffset], ul_Size );
				if ( advnvram_info->ul_IsPci == 0 )
				{
					outb( 0, advnvram_info->ul_BankEnableAddress);
				}
			}
			
			if ( i < ul_EndBank )
			{
				ul_DiskOffset = 0;
				ul_BufOffset += ul_Size;
				ul_Size       = len - ul_BufOffset;
				if ( ul_Size > advnvram_info->ul_BankLength )
					ul_Size = advnvram_info->ul_BankLength;
			}         
		} 
		if ( advnvram_info->ul_IsPci == 0 )
		{
			outb( 0, advnvram_info->ul_BankSwitchAddress);
		}  
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,31)	
	done:
		if (!__blk_end_request_cur(req, err))
			req = blk_fetch_request(q);
#else   
		end_request(req, TRUE );
#endif
	}
#else //2.4 kernel 
	while ( TRUE )
	{
		INIT_REQUEST;
		DEBUGPRINT( DEVICE_NAME": request %p: cmd %i sector %li (nr. %li) \n", 
		CURRENT, CURRENT->cmd, CURRENT->sector, CURRENT->current_nr_sectors );
		// check if the minor number is in range
		if ( DEVICE_NR( CURRENT->rq_dev ) > NUM_NVRAMDISKS )
		{
			INFOPRINT( DEVICE_NAME": request for unknown device.\n");
			end_request( 0 );
			continue;
		}
		minor = DEVICE_NR( CURRENT->rq_dev );
		advnvram_info  = &AdvNVRAMInfo[minor];

		// per sector = 512 Bytes
		start = CURRENT->sector << 9;    
		// the number of the sectors to transfer, per sector = 512 Bytes
		len  = CURRENT->current_nr_sectors << 9;  
		DEBUGPRINT(DEVICE_NAME": CURRENT->sector:%ld, current_nr_sectors:%ld, nr_sectors:%ld\n", 
		CURRENT->sector, CURRENT->current_nr_sectors, CURRENT->nr_sectors );

		ul_StartBank = start / advnvram_info->ul_BankLength;
		ul_EndBank   = ( start + len ) / advnvram_info->ul_BankLength;
		DEBUGPRINT( DEVICE_NAME": Start Bank: %ld, End Bank: %ld\n", 
		ul_StartBank, ul_EndBank );
      
		if ( ( start + len ) > advnvram_info->ul_DiskSize )
		{
			INFOPRINT( DEVICE_NAME ": bad access: block=%ld, count=%ld\n",
			CURRENT->sector,
			CURRENT->current_nr_sectors);
			INFOPRINT( DEVICE_NAME " : request past end of device\n");
			end_request( FALSE );
			continue;
		}	
      
		if ( ( CURRENT->cmd != READ ) && ( CURRENT->cmd != WRITE ) )
		{
			INFOPRINT( DEVICE_NAME ": bad command: %d\n", CURRENT->cmd );

			end_request( FALSE );

			continue;
		}

		ul_BufOffset  = 0;
		ul_DiskOffset = start % advnvram_info->ul_BankLength;
		ul_Size       = ( ul_DiskOffset + len <= advnvram_info->ul_BankLength )? 
						len : (advnvram_info->ul_BankLength - ul_DiskOffset );
      
		for ( i = ul_StartBank; i <= ul_EndBank; i++ )
		{
			uc_ptDisk = advnvram_info->uc_ptaddr + ul_DiskOffset;
			// set bank
			if ( advnvram_info->ul_IsPci == 0 )
			{
				outb( i, advnvram_info->ul_BankSwitchAddress );
			}
			if ( CURRENT->cmd == READ )
			{  
				DEBUGPRINT(DEVICE_NAME": READ-BANK = %d\n", i );
				memcpy( CURRENT->buffer, uc_ptDisk, ul_Size );            
			}
			else if ( CURRENT->cmd == WRITE )
			{
				DEBUGPRINT( DEVICE_NAME": WRITE-BANK = %d\n", i );
				if ( advnvram_info->ul_IsPci == 0 )
				{
					outb( 1, advnvram_info->ul_BankEnableAddress);
				}
				memcpy( uc_ptDisk, CURRENT->buffer, ul_Size );
				if ( advnvram_info->ul_IsPci == 0 )
				{
					outb( 0, advnvram_info->ul_BankEnableAddress);
				}
			}
      
			if ( i < ul_EndBank )
			{
				ul_DiskOffset = 0;
				ul_BufOffset += ul_Size;
				ul_Size       = len - ul_BufOffset;
				if ( ul_Size > advnvram_info->ul_BankLength )
				{
					ul_Size    = advnvram_info->ul_BankLength;
				}         
			}         
		} 
		if ( advnvram_info->ul_IsPci == 0 )
		{
			outb( 0, advnvram_info->ul_BankSwitchAddress);    
		}  

		end_request( TRUE );
	}
	
#endif 
	return ; 

}
