//-----------------------------------------------------------------
//  Demo program to test the NVRAM disk direct I/O
//  Advantech eAutomation Division
//  03-08-2013
//
//  Version: 1.0.0
//  Author:  WangMao
//  ChangeLog:
//  	--version 1.0.0  initial version.
//-----------------------------------------------------------------  

#include   <sys/types.h>
#include   <sys/stat.h>
#include   <unistd.h>
#include   <errno.h>
#include   <asm/fcntl.h>
#include   <string.h>
#include   <stdio.h>

//this macro enable the flag O_DIRECT when read/write file
#define __USE_GNU 1

#define BLOCK_SIZE  512
#define DATA_SIZE    BLOCK_SIZE

typedef char _aligned_char __attribute__ ((aligned (BLOCK_SIZE)));

char odata[DATA_SIZE] __attribute__ ((aligned (BLOCK_SIZE)));   // original data
char wdata[DATA_SIZE] __attribute__ ((aligned (BLOCK_SIZE)));
char rdata[DATA_SIZE] __attribute__ ((aligned (BLOCK_SIZE)));

int main ( int argc, char **argv )
{
   int i, cursor, count, nvramsize,nvramfd, retval=0;
   unsigned char command[50];

   FILE *pnvramfd;
   off_t ret;
  
   printf("To test %s\n",argv[1]); 
   //get the size of device
   sprintf( command, "blockdev --getsize %s", argv[ 1 ]);   
   pnvramfd = popen(command, "r");   
   if ( pnvramfd == NULL )
   {
      printf("POPEN error!\n");
      retval = 3;
      goto err;
   }
   if(fscanf(pnvramfd, "%d", &nvramsize) != 1)
   {
      printf("Error to get result from blockdev!\n");
      pclose(pnvramfd);
      retval = 4;
      goto err;
   }
   pclose( pnvramfd );
   
   /****************************************************************
    * please note the using of flag O_DIRECT which enable direct I/O
    ****************************************************************/
   if ((nvramfd = open( argv[ 1 ], O_RDWR | O_DIRECT )) == -1 )
   {
      printf("Error to open %s \n", argv[ 1 ]);
      retval = 2;
      goto out;
   } 

   
 
   /* prepare test pattern */
   for(i=0; i<DATA_SIZE; i++)
       wdata[i] = (i%2) ? 0x55 : 0xAA;

   // /* test direct I/O */
   for ( i = 0; i < nvramsize; i++ )
   {   
      printf("\rTesting direct I/O at sector %d", i);
      cursor = i * DATA_SIZE;
      if ( (ret = lseek( nvramfd, cursor, SEEK_SET )) < 0 )
      {
         printf("\nError to perform lseek: ret = %d", ret);
         retval = 5;
         goto err;
      }
   
      // backup nvram device data
      count = read( nvramfd, odata, sizeof(odata));
      if ( count != sizeof(odata))
      {
         printf("\nError to backup data, %d\n", count);
         retval = 6;
         goto err;
      }
      
      /* write test pattern to nvram */
      lseek( nvramfd, cursor, SEEK_SET );
      count = write( nvramfd, wdata, sizeof(wdata) );
      if ( count != sizeof( wdata ))
      {
         printf("\nError to write test pattern (%d, %d)\n",
                strlen(wdata), count);
         retval = 7;
         goto err;  
      }
 
      /* read back test pattern from nvram */     
      lseek( nvramfd, cursor, SEEK_SET );
      count = read( nvramfd, rdata, sizeof(rdata));
      if ( count != sizeof(rdata))
      {
         retval = 8;
         printf("\nError to read test pattern, %d\n", count);
         goto err;
      }

      /* compare test pattern */
      if ( !memcmp( rdata, wdata, count ) )
      {
         lseek( nvramfd, cursor, SEEK_SET );
         /* restore data to nvram */
         count = write( nvramfd, odata, sizeof(odata) );
         if ( count != sizeof( odata ))
         {
            printf("\nError to restore data (%d, %d)\n",
                   sizeof(odata), count);
            retval = 9;
            goto err;  
         }
         continue;
      }
      else
      {  
         printf("\nError!\n");
         retval = 10;
         goto err;
      }
   }
   printf("\nThe device is tested ok, NVRAM is ready for use!\n");

err:
   close( nvramfd );
out:
   return retval;
}
   
