//-----------------------------------------------------------------
//  Demo program to test the NVRAM disk via raw device
//  Advantech eAutomation Division
//  12-30-2004
//-----------------------------------------------------------------  

#include   <sys/types.h>
#include   <sys/stat.h>
#include   <unistd.h>
#include   <errno.h>
#include   <asm/fcntl.h>
#include   <string.h>
#include   <stdio.h>

#define BLOCK_SIZE      512
#define DATA_SIZE       BLOCK_SIZE

typedef char _aligned_char __attribute__ ((aligned (BLOCK_SIZE)));

//_aligned_char odata[DATA_SIZE];    // original data
//_aligned_char wdata[DATA_SIZE];
//_aligned_char rdata[DATA_SIZE];

char odata[DATA_SIZE]  __attribute__((aligned(BLOCK_SIZE)));
char wdata[DATA_SIZE]  __attribute__((aligned(BLOCK_SIZE)));
char rdata[DATA_SIZE]  __attribute__((aligned(BLOCK_SIZE)));

int main ( int argc, char **argv )
{
   int i, cursor, rawfd, count, nvramsize, retval=0;
   unsigned char command[50];
   FILE *nvramfd;
   off_t ret;

   if ( argc < 3 )
   {
      printf("Usage:\n");
      printf("%s /dev/raw/rawN /dev/NVRAMDRIVE \n", argv[ 0 ]);
      retval = 1;
      goto err;
   }
   
   if ((rawfd = open( argv[ 1 ], O_RDWR )) == -1 )
   {
      printf("Error to open %s \n", argv[ 1 ]);
      retval = 2;
      goto out;
   } 

   sprintf( command, "blockdev --getsize %s", argv[ 2 ]);   
   nvramfd = popen(command, "r");   
   if ( nvramfd == NULL )
   {
      printf("POPEN error!\n");
      retval = 3;
      goto err;
   }
   if(fscanf(nvramfd, "%d", &nvramsize) != 1)
   {
      printf("Error to get result from blockdev!\n");
      pclose(nvramfd);
      retval = 4;
      goto err;
   }
   
   //printf("The NVRAM size is: %d sectors.\n", nvramsize);
   pclose( nvramfd );
 
   /* prepare test pattern */
   for(i=0; i<DATA_SIZE; i++)
       wdata[i] = (i%2) ? 0x55 : 0xAA;

   /* test raw device */
   for ( i = 0; i < nvramsize; i++ )
   {   
      printf("\rTesting raw device at sector %d", i);
      cursor = i * DATA_SIZE;
      if ( (ret = lseek( rawfd, cursor, SEEK_SET )) < 0 )
      {
         printf("\nError to perform lseek: ret = %d", ret);
         retval = 5;
         goto err;
      }
   
      /* backup raw device data */
      count = read( rawfd, odata, sizeof(odata));
      if ( count != sizeof(odata))
      {
         printf("\nError to backup data, %d\n", count);
         retval = 6;
         goto err;
      }
      
      /* write test pattern to raw device */
      lseek( rawfd, cursor, SEEK_SET );
      count = write( rawfd, wdata, sizeof(wdata) );
      if ( count != sizeof( wdata ))
      {
         printf("\nError to write test pattern (%d, %d)\n",
                strlen(wdata), count);
         retval = 7;
         goto err;  
      }
 
      /* read back test pattern from raw device */     
      lseek( rawfd, cursor, SEEK_SET );
      count = read( rawfd, rdata, sizeof(rdata));
      if ( count != sizeof(rdata))
      {
         retval = 8;
         printf("\nError to read test pattern, %d\n", count);
         goto err;
      }

      /* compare test pattern */
      if ( !memcmp( rdata, wdata, count ) )
      {
         lseek( rawfd, cursor, SEEK_SET );
         /* restore data to raw device */
         count = write( rawfd, odata, sizeof(odata) );
         if ( count != sizeof( odata ))
         {
            printf("\nError to restore data (%d, %d)\n",
                   sizeof(odata), count);
            retval = 9;
            goto err;  
         }
         continue;
      }
      else
      {  
         printf("\nError!\n");
         retval = 10;
         goto err;
      }
   }
   printf("\nThe raw device is tested ok, NVRAM is ready for use!\n");

err:
   close( rawfd );
out:
   return retval;
}
   
